// gameflow_records.h
// Play Mechanix - Video Game System
// Copyright (c) 2010 Play Mechanix Inc. All Rights Reserved
//


#ifndef GAMEFLOW_RECORDS_H
#define GAMEFLOW_RECORDS_H


#include "game.h"
#include "game_score.h"


// player data that's reset at the start of each general file section
typedef struct GameFlow_PlayerGeneralRecord
{
	float timeRaw;
	float damageTaken;
	
	// extra health added (like from health pickups)
	float healthAdded;

} GameFlow_PlayerGeneralRecord;

// player data thats used to generate score
typedef struct GameFlow_PlayerScoreRecord
{
	// total number of bullets that hit an enemy
	int hits;

	// total number of bullets that missed an enemy
	int misses;

	// kills for each dino type
	int dino_kills[DINO_NUM_DINO];

	// number of circles we have killed
	int circle_kills;

	// streak info
	int streak_value;
	uns32 streak_start_tick;
	uns32 streak_end_tick;
	int streak_state; // 0 = no active streak, 1 = streaking

	// streak grace period
	uns32 streak_last_miss_tick;
	int streak_grace_miss_cnt;
	int streak_cnt;
	int streak_max;

	// count of bonus items hit
	int bonus_items[BONUS_NUM_BONUS];

	// pickups taken
	int pickupsShot[PICKUP_NUM_TYPES];
	
	float total;

}GameFlow_PlayerScoreRecord;

// player data that's reset at the start of each level
typedef struct GameFlow_PlayerLevelRecord
{
	int grenadesFired;

	int deaths;

	float damageTaken[DAMAGE_NUM_TYPES];
	// total of all types
	float damageTakenTotal;
	uns32 damageLastTime;

	// extra health added (like from health pickups)
	float healthAddedTotal;

	uns32 continueLastTime;


	// kills for each enemy type (dinos and anything else that can damage the player).
	// only counts enemies that are attacking the player
	int attackersKilled[ENEMY_NUM_TYPES];

	int lastHitTime;
	uns32 lastTrigInputTime;

	uns32 startTimeRaw;
	float timeRaw;

	uns32 creditsSpent;

	// keep this in the player level record so it gets reset with each level too
	GameFlow_PlayerGeneralRecord pgrec;

	// player data used in computing score
	GameFlow_PlayerScoreRecord score;
	// save the score at a checkpoint
	GameFlow_PlayerScoreRecord chkpt_score;

} GameFlow_PlayerLevelRecord;


// 7 + null char
#define PLAYER_MAX_NAME_SIZE	8


typedef struct GameFlow_PlayerRecord
{
	int pnum;

	char name[PLAYER_MAX_NAME_SIZE];

	float health;		// current player health
	int active;		// "active" means: alive, not on the continue screen
	int joined;		// "joined" means: alive or on the continue screen
	int deathsTotal;

	float maxHealth;							// full health for player

	GameFlow_PlayerLevelRecord plrec;

} GameFlow_PlayerRecord;


// data that's reset at the start of each general file section
typedef struct GameFlow_GeneralRecord
{
	float sec;
	// credits spent in this general section (based on damage taken)
	float dmgAsCred;
	
} GameFlow_GeneralRecord;


// game data that gets reset after each level
typedef struct GameFlow_LevelRecord
{
	int pickupsCreated[PICKUP_NUM_TYPES];

	// count enemies that have started attacking
	int attackersSpawned[ENEMY_NUM_TYPES];

	// maximum number of players that have been active at one time.
	int maxActive;

	// current general file (0 if none)
	int generalNum;

	int circles_tutorial_shown;

	// number of times the level was failed and restarted.
	int numFails;

	uns32 lastPlayerDeathTime;

	// keep this in the level record so it gets reset with each level too
	GameFlow_GeneralRecord ggrec;

} GameFlow_LevelRecord;


typedef struct GameFlow_GameRecord
{
	List players;	

	int levelStartTime;
	int levelEndTime;
	int levels_played[GAME_NUM_LEVELS];
	int acts_finished[GAME_NUM_LEVELS][GAME_ACTS_PER_LEVEL];
	int last_level_played;

	// minimum damage a player must take from each hit, no matter what the difficulty is
	float minDamage;

	// damage cooldowns
	uns32 damageCooldown[GAME_MAX_PLAYERS];		// min time between hits that will damage player
	uns32 joinCooldown;
	uns32 playerDeathCooldown;
	int damageCnt;

	float startDmgAdjustAfterDmg;
	float startDmgAdjustAfterTime;

	// difficulty value set in the adjustments menu
	float difficultyAdj;

	// affects how easy enemies are to kill. kept separate from overall difficulty.
	// clamped to range [-1,1]
	float enemyDifficulty;

	GameFlow_LevelRecord lrec;

} GameFlow_GameRecord;


void* MemAllocGameRecord(size_t size);

void GameFlow_GameRecord_Init();
void GameFlow_GameRecord_Free();
void GameFlow_GameRecord_RecordLevelAudits();

void GameFlow_LevelRecord_Init(GameFlow_LevelRecord* lrec);

void GameFlow_GameRecord_AddPlayer(int pnum);
void GameFlow_GameRecord_ContinuePlayer(int pnum);
int GameFlow_GameRecord_RemovePlayer(int pnum);
int GameFlow_GameRecord_UnjoinPlayer(int pnum);

// reset any records that don't persist across multiple levels
void GameFlow_Records_LevelReset();

// return global game record
GameFlow_GameRecord* GameFlow_Get_GameRecord();
// find player record by player number. return null if not found or not active
GameFlow_PlayerRecord* GameFlow_Get_PlayerRecord(int pnum);

// find player record by player number. return null if not found does not check active status
GameFlow_PlayerRecord* GameFlow_Get_PlayerRecordDisregardActiveStatus(int pnum);


// info helpers
int GameFlow_Get_ActiveCount();
int GameFlow_Get_JoinedCount();
int GameFlow_Get_PlayerRecordCount();
int GameFlow_Get_PlayerActive( int pnum );
void GameFlow_Set_PlayerActive( int pnum, int active );

float GameFlow_Get_DifficultyAdj();
float GameFlow_Get_EnemyDifficulty();
void GameFlow_Set_EnemyDifficulty(float diff);

// get the average seconds per credit for all players for the current level
float GameFlow_GetAvgTimePerCredLevel(int addHealth);
float GameFlow_GetAvgTimePerCredLevelPlyr(uns32 pnum, int addHealth);
// for the current general section
float GameFlow_GetAvgTimePerCredGeneral(int addHealth);

float GameFlow_Records_GetDamageTakenPlayer(int pnum, int addHealth);
float GameFlow_Records_GetGenDamageTakenPlayer(int pnum, int addHealth);

void GameFlow_UpdateActivePlayers();

// return 0 if still in cooldown, 1 otherwise
float GameFlow_DamagePlayer(GameFlow_PlayerRecord* prec, float damage, DamageType dmgType, int allowKill);

void GameFlow_RecordAttackerSpawn(EnemyType type);
// count an attacker kill
void GameFlow_RecordAttackerKill(uns32 pnum, EnemyType type);

void GameFlow_Records_SetGeneralNum(int num);

// tool callback for player mode
int GameFlow_Records_ToolPlayerMode( int nump );

// helper to create proper number of players when tool is running
void GameFlow_Records_CreateToolPlayers( void );

extern int gGameCabMaxPlayers;

#endif