// psm_engine.h
// Play Mechanix - Video Game System
// Copyright (c) 2007 Play Mechanix Inc. All Rights Reserved
//
// routines for building/using programmable state machines

#ifndef PSM_ENGINE_H
#define PSM_ENGINE_H

//#include <pmrt_utils/pmrt_utils.h>

#define PSM_MAX_NAME_LEN 32

#define PSM_PLAINTEXT_EVENTACTION "EVENTACTION"
#define PSM_PLAINTEXT_STATEMACHINEPARAM "SMPARAM"
#define PSM_PLAINTEXT_STATEMACHINENAME  "STATEMACHINE"


struct PSM_ExeContext; // forward dec

typedef struct PSM_State
{
	int number;
	char name[PSM_MAX_NAME_LEN];
	char params_type[RTDT_MAX_FORM_NAME_LEN];
	int  params_size;

	unsigned int flags;
	int (*StateFunc)(struct PSM_ExeContext *, void *, void *, uns32, float, float * );
	// run whenever the state exits
	int (*ExitFunc)(struct PSM_ExeContext *, void *, void * );

} PSM_State;


typedef struct PSM_StateRecord
{
	float elapsed_time;

	unsigned int updates;

	PSM_State *state;
	void	  *params;
	int		   params_size;

} PSM_StateRecord;


typedef struct PSM_ProgramInstruction
{
	char state_name[PSM_MAX_NAME_LEN];
	int  state_number;
	void *params;
	int  params_size;
} PSM_ProgramInstruction;



typedef struct PSM_ProgramEventAction
{
	int respond_once; // if set then drop event action after first time it is occurs 
	char event_name[PSM_MAX_NAME_LEN];
	PSM_ProgramInstruction instruction;

} PSM_ProgramEventAction;

typedef struct PSM_ProgramStateMachineParam
{
	char param_name[RTDT_MAX_FORM_NAME_LEN*4]; // give a little extra room to allow for array and sub-member setting
	char value[PSM_MAX_NAME_LEN];
} PSM_ProgramStateMachineParam;

typedef struct PSM_Program
{
	char name[PSM_MAX_NAME_LEN];
	char statemachine_name[PSM_MAX_NAME_LEN];

	List instructions;

	List eventactions;

	List statemachineparams;


	RTDataTypeParser *prtdtp;

	MemFuncs	mf;

} PSM_Program;






typedef struct PSM_StateMachine
{
	char name[PSM_MAX_NAME_LEN];
	char params_type[RTDT_MAX_FORM_NAME_LEN];
	int params_size;

	// run every time update is called on the statemachine
	int (*UpdateFunc)(struct PSM_ExeContext *, void *, uns32 );

	List states;

	RTDataTypeParser *prtdtp;

	MemFuncs	mf;

} PSM_StateMachine;



typedef struct PSM_LoopMarker
{
	int instruction_number; // what line number to return to
	int count; // how many times to loop to line_number
} PSM_LoopMarker;


typedef struct PSM_ExeContext
{


	PSM_StateMachine *statemachine;

	void	  *statemachine_params;

	PSM_Program      *program;

	int curr_instruction_number; // where we are in the current program

	uns32 instruction_count; // inc'd every time a  new instruction is loaded

	

	// StateFunc's are allowed to modify the passed in params and these changes are held over
	// until the ExeContext advances to the next instruction
	// hence we have a buffer that we copy the initial state params from our program into
	// before executing each instruction
	// note that calls to SetActiveInstruction(), will copy over this buffer with the new instruction's params
	// hence one must take care when using state_params w/in a state if it has called SetActiveInstruction()
	// buffer into which we put a copy of the state params for each instruction before it's executed	
	void			*state_params_buf; 

	float elapsed_time;

//	unsigned int last_update;
//	unsigned int update_elapsed_time;

	unsigned int updates;

	// data for most recent event
	void *event_data;

	RTDataTypeParser *prtdtp;




	// b/c the event actions may change
	// as the program runs, we need to maintain
	// a seperate list for each ExeContext
	// list is copied from program on init
	List eventactions;

	// put all the mem bufs we malloc for on a list
	// makes it easier to clean up
	List membufs;


	// list of instruction numbers to return to
	List return_stack;

	// list of instruction numbers to loop to
	List loop_stack;

	// The maximum number of continues that this
	// EXE is allowed to do in a row.
	int maxLoops;

	// current state data stored in here
	PSM_StateRecord  state_rec;

	// use this guy for keeping time if no time supplied
	PMRT_Timer timer;

	MemFuncs	mf;

	char logfile[PSM_MAX_NAME_LEN]; // if set then all instructions and event actions will be appended to log file as they are run...

} PSM_ExeContext;


typedef enum
{
	PSM_SUCCESS,
	PSM_TOOMANYLOOPS,
	PSM_INVALID_HANDLE,
	PSM_INVALID_PARAMS,
	PSM_MALLOC_FAILURE,
	PSM_LIST_FAILURE,
	PSM_FILE_FAILURE,
	PSM_FAILURE,
	PSM_PROGRAM_HALTED,
	PSM_PROGRAM_COMPLETED,
	PSM_INVALID_STATE,

} PSM_ERROR_CODES;


typedef enum
{
// timing
	PSM_BUILTIN_STATE_MSLEEP,
	PSM_BUILTIN_STATE_SLOOP,

// flow control
	PSM_BUILTIN_STATE_LABEL,
	PSM_BUILTIN_STATE_GOTO,
	PSM_BUILTIN_STATE_GOSUB,
	PSM_BUILTIN_STATE_RETURN,
	PSM_BUILTIN_STATE_INTERRUPT,
	PSM_BUILTIN_STATE_DO,
	PSM_BUILTIN_STATE_LOOP,
	PSM_BUILTIN_STATE_EXIT,

// variables
	PSM_BUILTIN_STATE_SET_PARAM,

// debug
	PSM_BUILTIN_STATE_PRINT,
	PSM_BUILTIN_STATE_SINGLECOMMENT,
	PSM_BUILTIN_STATE_STARTCOMMENT,
	PSM_BUILTIN_STATE_COMMENTLINE,
	PSM_BUILTIN_STATE_ENDCOMMENT,

	NUM_PSM_BUILTIN_STATES,
} PSM_BUILTIN_STATES;


typedef enum
{
	PSM_RETURN_WAIT,                // wait for next update call
	PSM_RETURN_CONTINUE,            // continue w/o waiting for next update call
	PSM_RETURN_NEXT_INSTRUCTION,    // proceed to next instuction w/o waiting for next update call
	PSM_RETURN_QUIT,                // stop program execution
} PSM_STATE_RETURN_CODES;

#ifdef __cplusplus
extern "C" {
#endif

int PSM_Engine_Init( RTDataTypeParser *prtdtp, void* (*malloc_func)(size_t), void  (*free_func) (void *) );


int PSM_State_Init( PSM_State *state, int number, char *name, char *params_type, int params_size, int (*StateFunc)(PSM_ExeContext *,void *,void *, uns32, float, float * ), int (*ExitFunc)(PSM_ExeContext *, void *, void *) );


int PSM_StateMachine_Init( PSM_StateMachine *statemachine, char *name, char *params_type, RTDataTypeParser *prtdtp, void* (*malloc_func)(size_t), void  (*free_func) (void *)  );
int PSM_StateMachine_AddBuiltInStates( PSM_StateMachine *statemachine );
int PSM_StateMachine_AddStates( PSM_StateMachine *statemachine, PSM_State *states, int count );
int PSM_StateMachine_SetUpdateFunc( PSM_StateMachine *statemachine, int (*UpdateFunc)(struct PSM_ExeContext *, void *, uns32 ) );

PSM_State *PSM_StateMachine_FindStateByName( PSM_StateMachine *statemachine, char *name );
int PSM_StateMachine_FindLargestParamSize( PSM_StateMachine *statemachine );

int PSM_StateMachine_DeInit( PSM_StateMachine *statemachine );
int PSM_StateMachine_Document( PSM_StateMachine *statemachine, char *filename, int show_types );

int PSM_Program_Init( PSM_Program *program, char *name, char *statemachine_name, void* (*malloc_func)(size_t), void  (*free_func) (void *) );
int PSM_Program_AddInstructions( PSM_Program *program, PSM_ProgramInstruction *instructions, int count, int start_pos );
int PSM_Program_DropInstructions( PSM_Program *program, int count, int start_pos );
PSM_ProgramInstruction *PSM_Program_GetInstruction( PSM_Program *program, int pos );
int PSM_Program_FindLabel( PSM_Program *program, char *label );


int PSM_Program_AddEventActions( PSM_Program *program, PSM_ProgramEventAction *eventactions, int count );
int PSM_Program_DropEventAction( PSM_Program *program, char *event_name );
PSM_ProgramEventAction *PSM_Program_GetEventAction( PSM_Program *program, char *event_name );

int PSM_Program_AddStateMachineParams( PSM_Program *program, PSM_ProgramStateMachineParam *statemachineparams, int count );
int PSM_Program_DropStateMachineParam( PSM_Program *program, char *param_name );
int PSM_Program_SetStateMachineParam( PSM_Program *program, char *param_name, char *value );
PSM_ProgramStateMachineParam *PSM_Program_GetStateMachineParam( PSM_Program *program, char *param_name );


int PSM_Program_DeInit( PSM_Program *program );

int PSM_Program_LoadFromPlainTextFile( PSM_Program *program, PSM_StateMachine *statemachine, char *filename );
int PSM_Program_LoadFromPlainTextBuf( PSM_Program *program, PSM_StateMachine *statemachine, char *filename );
int PSM_Program_SaveToPlainTextFile( PSM_Program *program, PSM_StateMachine *statemachine, char *filename );

int PSM_Program_CompileForStateMachine( PSM_Program *program, PSM_StateMachine *statemachine );


int PSM_ExeContext_Init( PSM_ExeContext *context, PSM_StateMachine *statemachine, void *statemachine_params, PSM_Program *program, void* (*malloc_func)(size_t), void  (*free_func) (void *) );
int PSM_ExeContext_Update( PSM_ExeContext *context, float *elapsed_time );
int PSM_ExeContext_DoEvent( PSM_ExeContext *context, char *event_name, void *event_data );
int PSM_ExeContext_DeInit( PSM_ExeContext *context );
int PSM_ExeContext_AddEventActions( PSM_ExeContext *context, PSM_ProgramEventAction *eventactions, int count );
int PSM_ExeContext_DropEventAction( PSM_ExeContext *context, char *event_name );
int PSM_ExeContext_DropAllEventActions( PSM_ExeContext *context );

PSM_ProgramEventAction *PSM_ExeContext_GetEventAction( PSM_ExeContext *context, char *event_name );
int PSM_ExeContext_SetActiveInstruction( PSM_ExeContext *context, PSM_ProgramInstruction *instruction );
int PSM_ExeContext_LoadCurrInstruction( PSM_ExeContext *context );

int PSM_ExeContext_SetMaxLoops( PSM_ExeContext *context, int maxLoops );

int PSM_BuiltInStateFunc_MSleep( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );
int PSM_BuiltInStateFunc_Sloop( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time, float *elapsed_time);

int PSM_BuiltInStateFunc_SingleComment( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );
int PSM_BuiltInStateFunc_StartComment( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );
int PSM_BuiltInStateFunc_EndComment( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );
int PSM_BuiltInStateFunc_Print( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time );

int PSM_BuiltInStateFunc_Goto( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );
int PSM_BuiltInStateFunc_GoSub( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );
int PSM_BuiltInStateFunc_Return( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );
int PSM_BuiltInStateFunc_Do( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );
int PSM_BuiltInStateFunc_Loop( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );

int PSM_BuiltInStateFunc_Interrupt( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );
int PSM_BuiltInStateFunc_SetParam( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );
int PSM_BuiltInStateFunc_Exit( PSM_ExeContext *context, void *statemachine_params, void *state_params, uns32 state_updates, float state_elapsed_time,  float *elapsed_time  );

#ifdef __cplusplus
}
#endif


#endif // PSM_ENGINE_H
