#ifndef QR_CODE_H_
#define QR_CODE_H_


#include "limits.h"

#include "g5.h"
#include "qrencode.h"


#define QR_MAX_URL_SIZE	4096

#define QR_CODE_MAX_ENTRIES	16

#define QR_ENTRY_MAX_BYTES	4096
#define QR_ENTRY_MAX_VALS	QR_ENTRY_MAX_BYTES


#define QR_ENTRY_SORT_KEY_MIN		(INT_MIN + 1)
#define QR_ENTRY_SORT_KEY_MAX		INT_MAX
#define QR_ENTRY_SORT_KEY_REMOVED	INT_MIN


typedef struct QREntry
{
	// size in bytes of each current value
	uns32 valSizes[QR_ENTRY_MAX_VALS];
	// offset for the beginning of each value
	uns32 valOffsets[QR_ENTRY_MAX_VALS];
	// number of values currently added
	uns32 numVals;
	// total bytes currently used to store values
	uns32 numBytes;

	// used to order entries. the smallest ones get kicked out of the QR
	// code when there are too many entries.
	int sortKey;

	// store values together within a single ASCII string.
	// value strings are only NULL-terminated if they don't fill their
	// whole char sequence.
	char vals[QR_ENTRY_MAX_BYTES];

} QREntry;


void QREntry_Init(QREntry* qre);

// return the total size in bytes of all values
uns32 QREntry_GetSize(QREntry* qre);
uns32 QREntry_GetNumVals(QREntry* qre);

char* QREntry_GetVal(QREntry* qre, uns32 valIdx);
// return size of value in bytes
uns32 QREntry_GetValSize(QREntry* qre, uns32 valIdx);
// return offset in bytes from the start of the value array
uns32 QREntry_GetValOffset(QREntry* qre, uns32 valIdx);

int QREntry_GetSortKey(QREntry* qre);
void QREntry_SetSortKey(QREntry* qre, int sortKey);

// append a new value to this entry. return index of the new value
uns32 QREntry_AddVal_Bytes(QREntry* qre, uns32 size, char* val);
uns32 QREntry_AddVal_Str(QREntry* qre, uns32 size, char* val);
uns32 QREntry_AddVal_Int(QREntry* qre, uns32 size, int val);
uns32 QREntry_AddVal_Flt(QREntry* qre, uns32 size, float val);

void QREntry_SetVal_Bytes(QREntry* qre, uns32 valIdx, char* val);
void QREntry_SetVal_Str(QREntry* qre, uns32 valIdx, char* val);
void QREntry_SetVal_Int(QREntry* qre, uns32 valIdx, int val);
void QREntry_SetVal_Flt(QREntry* qre, uns32 valIdx, float val);


typedef struct QRCodeData
{
	char urlStr[QR_MAX_URL_SIZE];

	uns32 nEntries;
	QREntry entries[QR_CODE_MAX_ENTRIES];

	// pointer to qrencode data
	QRcode* qrCode;

} QRCodeData;


void QR_Startup();
void QR_Shutdown();

void QR_Init(QRCodeData* qrd);
void QR_Free(QRCodeData* qrd);

void QR_SetURL(QRCodeData* qrd, char* urlStr);

int QR_AddEntry(QRCodeData* qrd, QREntry* qre);
QREntry* QR_EntryAt(QRCodeData* qrd, int idx);
// update the URL string with the current entries
void QR_EncodeToURL(QRCodeData* qrd);


typedef struct QRCodeImg
{
	// odr
	uns32 oid;
	uns32 texIdx;

	// width in pixels. equal to height
	float width;

} QRCodeImg;


void QR_NewImage( QRCodeData* qrd, QRCodeImg* qri, N2* link, float maxSize );
void QR_FreeImg( QRCodeImg* qri );


void QR_Test();


#endif