#ifndef RENDER_STATE_H
#define RENDER_STATE_H


// render_state.h
// Copyright 2011 Play Mechanix Inc. All Rights Reserved

#include "g5.h"

#define MAX_NUM_SHADMAP_SPLITS 3


#define M_OBJFRAMESTART_CLEARCLR 0x01
#define M_OBJFRAMESTART_CLEARZ   0x02
#define M_OBJFRAMESTART_CLEARSTENCIL 0x04



enum LightSourcesEnum
{
	SR_LIGHTSOURCE_NONE, // light came from nowheresville
	SR_LIGHTSOURCE_LIGHTGROUP, // light came from active light group
	SR_LIGHTSOURCE_GLOBAL, // light came from a global source
	SR_LIGHTSOURCE_OBJECT, // light came from an object

};


#define RENDERSTATE_MAX_LIGHTS 12

typedef struct RenderStateLight {
	int type;
	int src;
	// the next set of fields are packed tight
	// to make for more efficient transfer to shaders
	// specifically, 4 element of each array is borrowed, so comments on each for how
	float diffuse_color[4];  // a is linear attn
	float specular_color[4];  // a is quadratic attn
	float pos[4];  // pos in cam space (half vector for dir lights, assumes infinite view model), w is spot exp
	float dir[4];  // dir in cam space, w is spot_cutoff

} RenderStateLight;



typedef struct {
	int   on;
	// shader values
	// values laid out a tad funny to minimize uniform var count
	float uFogColor[4]; // 4th field is 1.0f - max_fog
	float uFogValues[2]; // 1st field is max_depth, 2nd field is 1/(max_depth-min_depth),
	// float uFogValuesOff[2]; // 1st field is max_depth, 2nd field is 1/(max_depth-min_depth), min_depth/max_depth set to camera far clip plane
} RenderStateFog;


#define RENDERSTATE_MAX_TEXS  9
#define RENDERSTATE_DRAW_BUFFERS  8

// Render State Tracking
// These structs serve two purposed
// one they track current OpenGL render
// state to prevent redundant OpenGL state changes
// second they track the 

// blend mode is abstracted slightly from GL
enum
{
	RENDERSTATE_BLEND_MODE_NONE,  // blend off
	RENDERSTATE_BLEND_MODE_ADD_SRC_ALPHA_AND_ONE_MINUS_SRC_ALPHA,
	RENDERSTATE_BLEND_MODE_ADD_SRC_ALPHA_AND_ONE,
	RENDERSTATE_BLEND_MODE_RSUB_SRC_ALPHA_AND_ONE_MINUS_SRC_ALPHA,
	RENDERSTATE_BLEND_MODE_RSUB_SRC_ALPHA_AND_ONE,
	RENDERSTATE_BLEND_MODE_CUSTOM,  // use this if you are setting a custom blend state not itemized above (consider add an enum and case in RenderState_SetBlendMode() for it)
};

// Cull Face is also abstracted
enum
{
	RENDERSTATE_CULL_FACE_NONE,  // cull face off
	RENDERSTATE_CULL_FACE_BACK,
	RENDERSTATE_CULL_FACE_FRONT,
	RENDERSTATE_CULL_FACE_RESTORE, // front or back, whichever we last set it to

};

// ditto alpha test
enum
{
	RENDERSTATE_ALPHA_TEST_NONE,  // off
	RENDERSTATE_ALPHA_TEST_GREATER_OR_EQUAL,
	RENDERSTATE_ALPHA_TEST_GREATER,
	RENDERSTATE_ALPHA_TEST_LESS_OR_EQUAL,
	RENDERSTATE_ALPHA_TEST_LESS,
};


#define RENDERSTATE_DRAWBUFFER0   0x00000001
#define RENDERSTATE_DRAWBUFFER1   0x00000002
#define RENDERSTATE_DRAWBUFFER2   0x00000004
#define RENDERSTATE_DRAWBUFFER3   0x00000008
#define RENDERSTATE_DRAWBUFFER4   0x00000010
#define RENDERSTATE_DRAWBUFFER5   0x00000020
#define RENDERSTATE_DRAWBUFFER6   0x00000040
#define RENDERSTATE_DRAWBUFFER7   0x00000080



#define RENDERSTATE_MAX_UNIFORM_BUFFER_BINDINGS  128

// binding points for engine supported uniform blocks

#define RENDERDATA_UNIFORM_BINDING_STD_MATRIX_DATA   3
#define RENDERDATA_UNIFORM_BINDING_STD_TEX_MATRIX_DATA   4
#define RENDERDATA_UNIFORM_BINDING_FOG_DATA          5
#define RENDERDATA_UNIFORM_BINDING_SHADMAP_DATA      6
#define RENDERDATA_UNIFORM_BINDING_LIGHT_DATA        7
#define RENDERDATA_UNIFORM_BINDING_SKEL_MATS         8
#define RENDERDATA_UNIFORM_BINDING_SKEL_MATS_PREV    9

typedef struct UniformDataBufferBinding
{
	Res *sudbc_res; // RESSUB_SHADERUNIFORMBLOCKBUFFERCOLLECTION that is bound to this bind point
	char buffer_name[PM_NAME_SIZE]; // key for sudbc_res's buffers hash whose val is the buffer object bound to this point
	uns32 udbo; // uniform buffer object currently bound to this binding
} UniformDataBufferBinding;


typedef struct {
	ObjCam *ocam;

	// these items are straight wrappers around the corresponding opengl state
	uns32  glMatrixMode;
	uns32  glActiveTexture;
	uns32  glClientActiveTexture;
	uns32  ShadeModel;

	uns32  AlphaTestMode;
	float  AlphaTestVal;

	// these modes are slightly abstracted for ease of use
	uns32  BlendMode;
	uns32  bBlendEnabled;

	uns32  CullFace;
	uns32  CullFaceRestore;

	uns32 bDepthWriteEnabled;
	uns32 bDepthTestEnabled;  
	uns32 DepthTestMode;  // depth func (GL_LEQUAL) etc...


	// which attrib arrays are currently enabled, generally set when shader is enabled
	uns32 VertAttribArrays;

	// uns32 glFrameBuffer;
	int DrawBuffersCount;
	uns32 DrawBuffers[RENDERSTATE_DRAW_BUFFERS];

	uns32 bTexActive[RENDERSTATE_MAX_TEXS];
	uns32 ActiveTexGen[RENDERSTATE_MAX_TEXS];
	uns32 LayerTexGen[RENDERSTATE_MAX_TEXS];
	uns32 ShaderTexGen[RENDERSTATE_MAX_TEXS];

	float Color[4];
	float ClearColor[4];
	float Material_SpecularShininess;
	float BloomThreshold;


	RenderStateFog Fog;


	ModelViewProjectionMatrix mvp_matrix;

	int TextureMatrixIsId[RENDERSTATE_MAX_TEXS];
	float TextureMatrix[RENDERSTATE_MAX_TEXS][M4_SIZE];


	// these items slightly higher level than general opengl state
	int  ShadMapsOn;
	float ShadIntensityGlobal;  // overriden by lg shad intensity if exists
	float uShadParams[3]; // r == Shad Intensity, g = ShadSplit1_Start, b = uShadSplit2_Start
	float uShadDir[3];
	float uShadDir_CamSpace[3];
	ModelViewProjectionMatrix shadmap_mvp_matrix[MAX_NUM_SHADMAP_SPLITS];

	int ActiveShaderIdx;
	Res *ActiveShader;
	ResDataShaderProgram *ActiveShaderRDS;  // taken from ActiveShader, guarunteed not to be null when ActiveShader != NULL (by RenderState_SetShader())
	// next three are copied from activeshader when it is set
	uns32 ActiveShaderFlags;
	uns32 ActiveShaderVertArrays;
	uns32 ActiveShaderMatrices; 

	//Uniform Block Data 
	int max_uniform_buffer_bindings; // min(RENDERSTATE_MAX_UNIFORM_BUFFER_BINDINGS, glGetIntegerv(GL_MAX_UNIFORM_BUFFER_BINDINGS))
	int max_vertex_shader_uniform_buffer_bindings;
	int max_geometry_shader_uniform_buffer_bindings;
	int max_fragment_shader_uniform_buffer_bindings;
	int max_uniform_block_size;

	UniformDataBufferBinding UniformDataBufferBindings[RENDERSTATE_MAX_UNIFORM_BUFFER_BINDINGS];

	RenderLayer  *ActiveRenderLayer;
	RenderList   *ActiveRenderList;
	RenderTarget *ActiveRenderTarget;


	

	// lights
	int ActiveLightGroup;
	char *ActiveLightProfile;
	float ActiveLightsByType[4];  // r = dirs, g = points, b = spots, a = unused
	void *LightGroupCurrentLightProfile[OBJ_MAX_LIGHT_GROUPS]; // if null, light group is not current w/ any light profile
	float LightGroupShadIntensity[OBJ_MAX_LIGHT_GROUPS];  // stashed so that we can restore it w/ restoring all light data
	RenderStateLight Lights[RENDERSTATE_MAX_LIGHTS];
	float gAmbient[4];
	int flashlightOn;
	float lightningMult;

	// lines and points
	float line_width;
	float point_size;

	// render the dynamic cube map from the position of this ODR
	uns32 dynCubeODR;

} RenderState;


extern RenderState gRenderState;


// RenderStateInit() must be called before core shaders are loaded
int RenderStateInit();
// RenderStateStart() must be called after core shaders are loaded
int RenderStateStart();

EXTERN_C void RenderState_SetMatrixMode(uns32 mode);
EXTERN_C void RenderState_SetActiveTexture(uns32 tex);
EXTERN_C void RenderState_SetClientActiveTexture(uns32 tex);

EXTERN_C void RenderState_BindTex( int index, int gltex, int tex);
EXTERN_C void RenderState_BindTexCube( int index, int gltex, int tex);
EXTERN_C void RenderState_UnBindTex( int tex );
EXTERN_C void RenderState_SetBlendMode( uns32 mode );
EXTERN_C void RenderState_SetCullFace( uns32 mode );
EXTERN_C void RenderState_SetDepthTest( int on, uns32 mode );
EXTERN_C void RenderState_SetDepthWrite( int on );

EXTERN_C void RenderState_SetColor( float r, float g, float b, float a );
EXTERN_C void RenderState_SetColorv( float *rgba );
EXTERN_C void RenderState_SetClearColor( float *rgba );

EXTERN_C void RenderState_SetDrawBuffers( int count, uns32 *buffers );
EXTERN_C void RenderState_SetDrawBuffersSimple( int count ); // assumes you want standard ordering


EXTERN_C void RenderState_SetFogColor(float *rgba); // a is 1 - max fog
EXTERN_C void RenderState_SetFogDepth(float min_depth, float max_depth);
EXTERN_C void RenderState_SetFogOn(int on);

//EXTERN_C void RenderState_SetFrameBuffer( uns32 fbo );


EXTERN_C void RenderState_SetProjectionMatrix( float *m4 );
EXTERN_C void RenderState_SetViewMatrix( float *m4 );
EXTERN_C void RenderState_SetViewProjMatrices( ObjCam *ocam );
EXTERN_C void RenderState_SetModelMatrix( float *m4 );
EXTERN_C void RenderState_SetModelMatrixByVM( float *vm );
EXTERN_C void RenderState_SetModelMatrixByTransRotScale( float *trans, float *rot, float *scale );
EXTERN_C void RenderState_SetModelMatrixByRD( ObjRenderData* rd );
EXTERN_C void RenderState_SetupBasic2DCamera(  float start_x, float start_y, float w, float h );


EXTERN_C void RenderState_ShaderSendModelViewProjectionAndTextureMatrices();

EXTERN_C void RenderState_SetTextureMatrix( int tex, float *m4 );
EXTERN_C void RenderState_SetTextureMatrixbyRD( int tex, ObjRenderData *rd );


EXTERN_C void RenderState_SetShadMaps( int on );
EXTERN_C void RenderState_SetShadDir( float *vec );
EXTERN_C void RenderState_SetShadIntensityGlobal( float intensity );
EXTERN_C void RenderState_SetShadIntensity( float intensity );
EXTERN_C void RenderState_SetShadSplits( float Split1_Start, float Split2_Start );
EXTERN_C void RenderState_SetShadMapProjMatrix( int shad_split, float *m4 );
EXTERN_C void RenderState_SetShadMapViewMatrix( int shad_split, float *m4 );
EXTERN_C void RenderState_ShadMapDataBlockCommit();

EXTERN_C void RenderState_SetShader( Res *shader, uns32 renderflags );
EXTERN_C void RenderState_SetCoreShader( int index, uns32 renderflags );  // index is from CoreShadersEnum
EXTERN_C void RenderState_SetEnabledVertAttribArrays( uns32 arrays );
EXTERN_C void RenderState_EnableVertAttribArrays( uns32 arrays );
EXTERN_C void RenderState_DisableVertAttribArrays( uns32 arrays );
EXTERN_C void RenderState_SendRenderDataColor( ObjRenderData* rd );


EXTERN_C void RenderState_SetRenderTarget(RenderTarget *rt, int screen);
EXTERN_C void RenderState_SetRenderLayer(RenderLayer *rlayer);


EXTERN_C int  RenderState_SetActiveLightGroup( int lightgroup );
//EXTERN_C int RenderState_TurnOffLightsBySource( char src );
//EXTERN_C int RenderState_TurnOnObjLights( Obj *obj );
//EXTERN_C void RenderState_TurnOffLights();

EXTERN_C int RenderState_BindUniformBlockBuffer( int bind_point, char *block_name, char *buffer_name );
EXTERN_C int RenderState_UnBindUniformBlockBuffer( char *block_name, char *buffer );

EXTERN_C void RenderState_SetSpecularExponent( float val );
EXTERN_C void RenderState_SetBloomThreshold( ObjRenderData* rd );
EXTERN_C void RenderState_SetLineWidth( float line_width );


#endif

