// sound.h
// Play Mechanix - Video Game System
// Copyright (c) 2010 Play Mechanix Inc. All Rights Reserved

#ifndef SOUND_H
#define SOUND_H

#define ENABLE_FMOD		(0)	// for surround sound (LICENSE: NOT FREE!)
#define ENABLE_OPENAL	(0)	

#if (ENABLE_FMOD==0 && ENABLE_OPENAL==0)
#define USE_SDL		(1)
#include "sound_sdl.h"
#define SOUND_MONO_PLUS_RUMBLE (0) // push all sounds to one stereo channel and rumble sounds to the other
#elif ENABLE_OPENAL
#define SOUND_MONO_PLUS_RUMBLE (0) 
#define USE_OPENAL	(1)
#include "sound_openal.h"
#else
#define USE_FMOD	(1)
#include "sound_fmod.h"
#define SOUND_MONO_PLUS_RUMBLE (0)
#endif



// Sound channels, uns32 for combinations
#define SND_CH0		0x00000001						
#define SND_CH1		0x00000002
#define SND_CH2		0x00000004
#define SND_CH3		0x00000008
#define SND_CH4		0x00000010
#define SND_CH5		0x00000020
#define SND_CH6		0x00000040
#define SND_CH7		0x00000080
#define SND_CH8		0x00000100
#define SND_CH9		0x00000200
#define SND_CH10	0x00000400
#define SND_CH11	0x00000800
#define SND_CH12	0x00001000
#define SND_CH13	0x00002000
#define SND_CH14	0x00004000
#define SND_CH15	0x00008000
#define SND_CH16	0x00010000
#define SND_CH17	0x00020000
#define SND_CH18	0x00040000
#define SND_CH19	0x00080000
#define SND_CH20	0x00100000
#define SND_CH21	0x00200000
#define SND_CH22	0x00400000
#define SND_CH23	0x00800000
#define SND_CH24	0x01000000
#define SND_CH25	0x02000000
#define SND_CH26	0x04000000
#define SND_CH27	0x08000000
#define SND_CH28	0x10000000
#define SND_CH29	0x20000000
#define SND_CH30	0x40000000
#define SND_CH31	0x80000000
#define SND_CHALL 	0xFFFFFFFF

#define SND_CHANNEL_NUM		32

// Priority
#define SND_PRI_HIGH_NO_STOP    512
#define SND_PRI_HIGH			256
#define SND_PRI_MED				128
#define SND_PRI_LOW				16

// Sound ID
#define MAX_ID		    2147483647

// Volume define, float volume
#define MAX_VOLUME      255.0f
#define MIN_VOLUME	    0.0f

// Panning define, float panning pct
#define MAX_PAN		100.0f
#define MIN_PAN		0.0f

#define SND_CHFLAG_BGM			0x00000001	// this channel is playing back ground music
#define SND_CHFLAG_DUCK_INDF	0x00000002	// this channel has been ducked indefinitely
#define SND_CHFLAG_BLEND_RAW	0x00000004	// use raw timer for blending



// AttenuateRange, z for total volume, left and right for panning effect
typedef struct AttenuateRange
{
	float zMin;				// Total attenuate
	float zMax;					
	float left;					
	float right;
	float rolloff;
	float pan_ratio; // how much of sound vol is subject to panning (rest is always centered)
} AttenuateRange;

// Sound system setup
typedef struct SoundSysSetup
{
	int		stereo;									// Stereo?? should be 2 or 1
	int		channels;								// By default, 16
	int		frequency;								// By default, 22050
	int		buffersize;								// Default buffer size for game, 4096
	uns16	format;									// Default format: AUDIO_S16SYS
	int		log;
} SoundSysSetup;

// Play list, for random play
typedef struct PlayList
{
	char		name[64];							// Name of the play list
	List		list;								// List of the strings
	uns32		last_played;						// index of the last sound played
} PlayList;

// Single sound track
typedef struct Sound
{
	char		name[64];							// Name of the sound
	SoundPtrType	data;								// Sound Data
} Sound;


// Sound bank
typedef struct SoundBank 
{
	char		name[64];							// Name of the sound bank
	HashTable	sounds;								// Hash table for all sounds in this bank
	HashTable	playLists;							// Hash table for the play list
	HashTable	dfltVolumes;						// Hash table for the default volume to play sounds at
	uns32		mark;								// res mark, sound banks w/o a mark are purged on wipeout
} SoundBank;

// SingleChannel
// Contains all information in the current channel
typedef struct SingleChannel
{
	int32	id;										// Unique play id for the sound, if id == -1, it is empty
	int		done;									// sound is done playing (set by mixer thread when sound has completed)
	float	volume;									// Volume of this channel
	float   current_volume;							// Attenuated volume
	int		priority;								// Priority of the playing music
	uns32	mask;									// Channel mask
	int		loop;									// Loop how many times, 0 = no loop, just play once
	int		age;									// Age of the channel
	int		flags;									// Any active channel flags

	timed_blend blend;								// For time blending

	float duck_start_vol;							// Ducking a sound
	uns32 duck_time;
	uns32 duck_tick;

	// For position
	float	world_pos[3];							// World position of this channel
	int		attenuate;								// Whether we should attenuate this sound with dis
	int		has_pos;								// Whether the world_position data is valid
	float	rolloff;

	int		has_obj;								// Whether this one has an object
	uns32   objid;									// Binded object id
	uns32   camid;									// Binded camera id
	uns32	expire_tick;							// have a sound stop playing after a period of time
	uns32	expire_time;

	float masterVolOverride;

	// used by SDL. music gets streamed from disk
	int is_music_sdl;


	SoundBank* bank;								// Binded sound bank					
	Sound* sound;									// Binded sound 
} SingleChannel;

// For quick search
typedef struct SoundSearch
{
	char  name[64];															
	Sound* sound;
} SoundSearch;

////////////////////////////////////////////////// Globals //////////////////////////////////////////////////

extern SoundSysSetup	gSoundParameter;
extern AttenuateRange	gAttenuateRange;
extern float			gMasterVolume;
extern int32			gPlayingUniqueId;

extern List				gBankList;
extern SingleChannel*	gChannels;	
extern int				gEnableRumble;
extern int				gRumbleVolume;
extern int				gRumbleVolume_minigun;
extern int				gRumbleVolume_electro;
extern int				gRumbleVolume_freeze;
extern int				gRumbleVolume_shotgun;
extern int				gRumbleVolume_hit;

////////////////////////////////////////////////// Functions //////////////////////////////////////////////////

// System setup
void SoundSetup(int stereo, int channels, int frequency, int buffersize, uns16 format);		// Setup sound system
void SoundSetRange(float minDepth, float maxDepth, float left, float right, float rolloff, float pan_ratio );									// Set attenuate range
void SoundSetMasterVolume(float master);													// Set the master volume
float SoundGetMasterVolume(void);															// Get the master volumes
void SoundOverrideMasterVolume( int32 id, float masterVol );

// Loop
int  SoundInit();																			// Initialize SDL_Mixer, allocate channels, 16 channels by default
int  SoundLoop();																			// Update the channel
int  SoundWipeout();																		// Unload all banks w/ a blank mark
void SoundFinal();																			// Shut down all stuffs about sound

// Sound bank
int  SoundLoadBank(char* bankname,uns32 mark);												// Load a sound bank, add a sound bank in the hash table
int  SoundUnloadBank(char* bankname);														// Unload a sound bank
int  SoundReloadBank(char* bankname);														// Reload a sound bank	
int  SoundBankClearMark(char* bankname,uns32 mark);											// clears passed mark bits from sound bank
int  SoundBanksClearMark(uns32 mark);														// clears passed mark bits from all sound bank

// Sound play
int32 SoundPlay(char* soundname, uns32 channel, float volume, int loop, int priority);										// Play a single sound, on which channel
int32 SoundPlayWithPos(char* soundname, uns32 channel, float volume, int loop, int priority, float x, float y, float z);	// Play sound by position
int32 SoundPlayWithPosNewRolloff(char* soundname, uns32 channel, float volume, int loop, int priority, float x, float y, float z, float rolloff );
int32 SoundPlayWithObj(char* soundname, uns32 channel, float volume, int loop, int priority, uns32 objid, int camid );      // Play sound w/ position taken from objid
int32 SoundPlayWithObjNewRolloff(char* soundname, uns32 channel, float volume, int loop, int priority, uns32 objid, int camid, float rolloff );
int32 SoundPlayRumble(char* soundname, uns32 channel, float volume, int loop, int priority);										// Play a single sound, on which channel
//void SoundSetRumbleById(uns32 id, int rumble);

// Sound stop
void SoundStopByChannel(uns32 channelmask);													// Stop playing a sound by channel
void SoundStopById(uns32 id);																// Stop playing a sound by id
void SoundStopByName(char* soundname);														// Stop playing a sound by name
void SoundStopAll();																		// Stop playing all channels
void SoundExpireById(uns32 id, uns32 ms);													// Stop playing sound after designated time
void SoundExpireByName(char* soundname, uns32 ms);
void SoundStopRumbleById(int32 id);
void SoundStopRumbleByName(char* name);

// Sound check if playing
int SoundIsPlayingByChannel(uns32 channelmask);
int SoundIsPlayingById(uns32 id);
int SoundIsPlayingByName(char* soundname);
int SoundIsPlayingByNameAndOID(char* soundname, uns32 objid);




// Sound adjust volume
void SoundAdjustVolumeByChannel(uns32 channelmask, float volume, uns32 ms);					// Adjust volume by channel
void SoundAdjustVolumeById(uns32 id, float volume, uns32 ms);								// Adjust volume by id
void SoundAdjustVolumeByName(char* soundname, float volume, uns32 ms);						// Adjust volume by name

// Sound ducking
void SoundDuckByName(char* soundname, float volume, uns32 time_ms, uns32 blend_ms);
void SoundDuckAllBackGroundMusic( uns32 fade_out_time );									// Mutes all sounds not as background music
void SoundResumeAllBackGroundMusic( uns32 fade_in_time );								// Unmute all sounds not as background music
void SoundDuckAllNonBackGroundMusic( uns32 fade_out_time );									// Mutes all sounds not marked as background music
void SoundResumeAllNonBackGroundMusic( uns32 fade_in_time );								// Unmute all sounds not marked as background music
void SoundDuckAllSounds( uns32 fade_out_time );												// Mutes all sounds
void SoundResumeAllSounds( uns32 fade_in_time );											// Unmute all sounds


// Sound pause
void SoundPauseByName(char* soundname);
void SoundResumeByName(char* soundname);

void SoundSetBGMById( uns32 id );															// Marks a sound as background music

void SoundAddChannelFlagByID( uns32 id, uns32 flag );
void SoundRemoveChannelFlagByID( uns32 id, uns32 flag );


// Sound attenuate, panning
int  SoundPanning(uns32 channel, float left_vol);												// Start sound panning effect on a single channel
int  SoundPanningByName(char* soundname, float left_vol);
int  SoundPanningById(uns32 id, float left_vol);


void SoundSetPos(uns32 channel, float x, float y, float z);									// Set pos of this sound, enable attenuate
void SoundSetPosByName(char* soundname, float x, float y, float z);
void SoundSetPosById(uns32 id, float x, float y, float z);


void SoundTurnOffAttenuate(uns32 channelmask);												// Turn off attenuate, turn off panning
void SoundTurnOffAttenuateByName(char* soundname);
void SoundTurnOffAttenuateById(uns32 id);


void SoundTurnOnAttenuate(uns32 channelmask);												// Turn on attenuate, turn on panning
void SoundTurnOnAttenuateByName(char* soundname);
void SoundTurnOnAttenuateById(uns32 id);

// Bind camera and bind object
//void SoundBindCamera(uns32 channel, ObjCam* cam);											// Bind a camera 
//void SoundBindObj(uns32 channel, Obj* obj);												// Bind an object to this sound channel
void SoundBindCamera(uns32 channel, uns32 camid);											// Bind a camera 
void SoundBindCameraByName(char* soundname, uns32 camid);
void SoundBindCameraById(uns32 id, uns32 camid);

void SoundBindObj(uns32 channel, uns32 objid);												// Bind an object to this sound channel
void SoundBindObjByName(char* soundname, uns32 objid);
void SoundBindObjById(uns32 id, uns32 objid);

void SoundSetSplineByName(char* soundname, char* splineResName, int cam_num, float xOff, float yOff, float zOff);	//Create spline and bind


int SoundGetByObj(uns32 objid, int match);									// find sound id of sound number 'match' bound to this object


typedef void (SND_LIST_NAMES_CB)(char *, int);

int SoundListNames( SND_LIST_NAMES_CB (*func) );
int SoundListBankNames( SND_LIST_NAMES_CB func );
int SoundListNamesInBank( char* bankName, SND_LIST_NAMES_CB func );
int SoundIsBankName( char* name );

// For debug
char* ShowChannelInfo();
char* ShowBankInfo();

////////////////////////////////////////////////// Helpers //////////////////////////////////////////////////
void SoundLogStatus();
void ClearChannelInfo(int i);
int Power2(int i);

///////////////////////////////////////////// Utilization Log ///////////////////////////////////////////////
void SoundLogCreate();
void SoundLogDestroy();
void SoundLogDiff();
void SoundLogFile( char* name );
void SoundLogPlaylist( char* name );


///////////////////////////////////// Defined in lib specific files /////////////////////////////////////////
void FreeSoundData(Sound* sound);
void LoadSoundData(SoundPtrType *data, char *filename);
void SoundLibUpdate();
void SoundLibUpdateLoc(int channel);
void SoundLibSetLoc(int channel, float x, float y, float z);
void SoundLibSetChannelVolume(int channel, float volume);
void SoundLibExpireChannel(int channel, int ms);
void SoundLibStopChannel(int channel);
void SoundLibStopAll();
void SoundLibShutdown();
int SoundLibVolumeMap(float volume);
int SoundLibSetPanVolume(int channel, unsigned char left_volume, unsigned char right_volume);
// non-linear
int SoundLibSetPanVolumeNL(int channel, unsigned char left_volume);
int SoundLibPlaySoundOnChannel(int channel, Sound* sound, int do_loop);
int SoundLibSetRumble(int channel, int rumble);
void SoundLibPauseSoundOnChannel(int channel);
void SoundLibResumeSoundOnChannel(int channel);

#endif // SOUND_H

